/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozCmgui.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "ComputedFieldZincXpidl.hpp"
#include "NodeZincXpidl.hpp"
#include "SceneViewerZincXpidl.hpp"
#include "TimeSequenceZincXpidl.hpp"
#include "CommandDataZincXpidl.hpp"
extern "C"
{
#include <command/cmiss.h>
#include <api/cmiss_time_sequence.h>
#include <api/cmiss_core.h>
#include <image_processing/computed_field_thresholdFilter.h>
#include <computed_field/computed_field_component_operations.h>
#include <computed_field/computed_field_window_projection.h>
}
#include "nsXPCOM.h"

CmguiComputedField::CmguiComputedField()
{
  /* constructor code */
}

CmguiComputedField::~CmguiComputedField()
{
  /* destructor code */
}

/* void evaluateAtNode (in CmguiINode n, in double time, [array, size_is (sizeof_retval)] out double retval, out PRUint32 sizeof_retval); */
NS_IMETHODIMP CmguiComputedField::EvaluateAtNode(CmguiINode *n, double time, double **retval, PRUint32 *sizeof_retval)
{
  CmguiNode *n_cast = NULL;
  if ((NULL == n)
    || (NS_OK != n->QueryInterface(NS_GET_IID(CmguiNode), (void **)&n_cast))
    || (NULL == n_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }
  unsigned int i, number_of_components = Cmiss_computed_field_get_number_of_components(impl);
  float *values = (float *)NS_Alloc(sizeof(float) * number_of_components);

  if (!(Cmiss_computed_field_evaluate_at_node(impl,
			  n_cast->impl, time, number_of_components, values)))
  {
          NS_Free(values);
	  return NS_ERROR_UNEXPECTED;
  }
  *sizeof_retval = number_of_components;
  *retval = NS_STATIC_CAST(double*,NS_Alloc(sizeof(double) * number_of_components));
	for (i = 0 ; i < number_of_components ; i++)
	{
		(*retval)[i] = values[i];
	}
  NS_Free(values);
  return NS_OK;
}

/* void setValuesAtNode (in CmguiINode n, in double time, [array, size_is (sizeof_values)] in double values, in PRUint32 sizeof_values); */
NS_IMETHODIMP CmguiComputedField::SetValuesAtNode(CmguiINode *n, double time, double *values, PRUint32 sizeof_values)
{
  CmguiNode *n_cast = NULL;
  if ((NULL == n)
    || (NS_OK != n->QueryInterface(NS_GET_IID(CmguiNode), (void **)&n_cast))
    || (NULL == n_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }
  float *float_values = (float *)NS_Alloc(sizeof(float) * sizeof_values);
	unsigned int i;
	for (i = 0 ; i < sizeof_values ; i++)
	{
		float_values[i] = values[i];
	}

	if (!(Cmiss_computed_field_set_values_at_node(impl,
				n_cast->impl, time, sizeof_values, float_values)))
	{
 	        NS_Free(float_values);
		return NS_ERROR_UNEXPECTED;
	}
  NS_Free(float_values);
  return NS_OK;
}

/* void setStringAtNode (in long componentNumber, in CmguiINode n, in double time, in string string_value); */
NS_IMETHODIMP CmguiComputedField::SetStringAtNode(PRInt32 componentNumber, CmguiINode *n, double time, const char *string_value)
{
  CmguiNode *n_cast = NULL;
  if ((NULL == n)
    || (NS_OK != n->QueryInterface(NS_GET_IID(CmguiNode), (void **)&n_cast))
    || (NULL == n_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }
  if (!Cmiss_computed_field_is_type_finite_element(impl))
  {
    return NS_ERROR_UNEXPECTED;
  }

	if (!(Cmiss_computed_field_finite_element_set_string_at_node(
		impl, componentNumber, n_cast->impl, time, string_value)))
	{
		return NS_ERROR_UNEXPECTED;
	}
  return NS_OK;
}

/* void defineAtNode (in CmguiINode n, in CmguiITimeSequence timeSequenceIn, in CmguiINodeFieldCreator nodeFieldCreatorIn); */
NS_IMETHODIMP CmguiComputedField::DefineAtNode(CmguiINode *n, CmguiITimeSequence *timeSequenceIn, CmguiINodeFieldCreator *nodeFieldCreatorIn)
{
  CmguiNode *n_cast = NULL;
  if ((NULL == n)
    || (NS_OK != n->QueryInterface(NS_GET_IID(CmguiNode), (void **)&n_cast))
    || (NULL == n_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }

  /* Optional parameters */
  Cmiss_time_sequence_id time_sequence = NULL;
  Cmiss_node_field_creator_id creator = NULL;

  CmguiTimeSequence *timeSequenceIn_cast = NULL;
  if (timeSequenceIn 
	  && (NS_OK == timeSequenceIn->QueryInterface(NS_GET_IID(CmguiTimeSequence), (void **)&timeSequenceIn_cast))
	  && timeSequenceIn_cast)
  {
	  time_sequence = timeSequenceIn_cast->impl;
  }
  CmguiNodeFieldCreator *nodeFieldCreatorIn_cast = NULL;
  if (nodeFieldCreatorIn 
	  && (NS_OK == nodeFieldCreatorIn->QueryInterface(NS_GET_IID(CmguiNodeFieldCreator), (void **)&nodeFieldCreatorIn_cast))
	  && nodeFieldCreatorIn_cast)
  {
	  creator = nodeFieldCreatorIn_cast->impl;
  }

  if (!Cmiss_computed_field_is_type_finite_element(impl))
  {
    return NS_ERROR_UNEXPECTED;
  }
  if (!(Cmiss_computed_field_finite_element_define_at_node(impl,
	  n_cast->impl, time_sequence, creator)))
  {
    return NS_ERROR_UNEXPECTED;
  }
  return NS_OK;
}


/* 	[cmguihandcoded] void getTypeConnectedThresholdImageFilter(out CmguiIComputedField source_field, 
																						out double lower_threshold, 
																						out double upper_threshold,
																						out double replace_value,
																						[array, size_is(dimension)] out double seed, out PRUint32 dimension);
*/
NS_IMETHODIMP CmguiComputedField::GetTypeConnectedThresholdImageFilter(CmguiIComputedField **source_field, double *lower_threshold,
																																			 double *upper_threshold, double *replace_value, 
																																			 PRUint32 *retnum_seed_points, PRUint32 *retdimension,
																																			 double **retseedValues, PRUint32 *sizeof_retseedValues)
{

  unsigned int i;
  Cmiss_computed_field_id source_field_cmgui;
	int num_seed_points;
	int dimension;

	double *seedValues; 

  if (!(Cmiss_computed_field_get_type_connected_threshold_image_filter(impl,&source_field_cmgui,lower_threshold,upper_threshold,replace_value,&num_seed_points,&dimension,&seedValues)))
  {
		Cmiss_deallocate(seedValues);
	  return NS_ERROR_UNEXPECTED;
  }

	*retnum_seed_points = (unsigned int)num_seed_points;
	*retdimension = (unsigned int)dimension;

  *source_field = new CmguiComputedField(source_field_cmgui);
  NS_ADDREF(*source_field);

  *sizeof_retseedValues = num_seed_points * dimension;
  *retseedValues = NS_STATIC_CAST(double*,NS_Alloc(sizeof(double) * *sizeof_retseedValues));
	for (i = 0 ; i < *sizeof_retseedValues; i++)
	{
		(*retseedValues)[i] = seedValues[i];
	}
	
	Cmiss_deallocate(seedValues);
  return NS_OK;
}


/*	[cmguihandcoded] void setTypeConnectedThresholdImageFilter(in CmguiIComputedField source_field, 
																						in double lower_threshold, 
																						in double upper_threshold,
																						in double replace_value,
																						[array, size_is(dimension)] in double seed, in PRUint32 dimension);
**/
NS_IMETHODIMP CmguiComputedField::SetTypeConnectedThresholdImageFilter(CmguiIComputedField *source_field, double lower_threshold, double upper_threshold, double replace_value, PRUint32 num_seed_points, PRUint32 dimension, double *seedValues, PRUint32 sizeof_seedValues)
{


  CmguiComputedField *source_field_cast = NULL;
  if ((NULL == source_field)
    || (NS_OK != source_field->QueryInterface(NS_GET_IID(CmguiComputedField), (void **)&source_field_cast))
    || (NULL == source_field_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }

	// pjb: is this necessary ?
  double *double_seed_values = (double *)NS_Alloc(sizeof(double) * sizeof_seedValues);
	unsigned int i;
	for (i = 0 ; i < sizeof_seedValues ; i++)
	{
		double_seed_values[i] = seedValues[i];
	}
	
	if ( !(Cmiss_computed_field_set_type_connected_threshold_image_filter(impl, source_field_cast->impl, lower_threshold, upper_threshold, replace_value, num_seed_points, dimension, double_seed_values)) )
	{
		NS_Free(double_seed_values);
		return NS_ERROR_UNEXPECTED;
	}

	NS_Free(double_seed_values);
  return NS_OK;
	
}


/* void getTypeSumComponents(out CmguiIComputedField source_field, [array, size_is(count)] out double weights, out PRUint32 count); */
NS_IMETHODIMP CmguiComputedField::GetTypeSumComponents(CmguiIComputedField **source_field, double **retweights, PRUint32 *sizeof_retweights)
{

  unsigned int i;

  Cmiss_computed_field_id source_field_cmgui;

	float *weights; 

  if (!(Cmiss_computed_field_get_type_sum_components(impl,&source_field_cmgui, &weights)))
  {
		Cmiss_deallocate(weights);
	  return NS_ERROR_UNEXPECTED;
  }

  *source_field = new CmguiComputedField(source_field_cmgui);
  NS_ADDREF(*source_field);

	unsigned int number_of_components = Cmiss_computed_field_get_number_of_components(source_field_cmgui);	

  *sizeof_retweights = number_of_components;
  *retweights = NS_STATIC_CAST(double*,NS_Alloc(sizeof(double) * number_of_components));
	for (i = 0 ; i < number_of_components ; i++)
	{
		(*retweights)[i] = weights[i];
	}
	
	Cmiss_deallocate(weights);
  return NS_OK;
}


/* 	void setTypeSumComponents(in CmguiIComputedField source_field, [array, size_is(count)] in double weights, in PRUint32 count); */
NS_IMETHODIMP CmguiComputedField::SetTypeSumComponents(CmguiIComputedField *source_field, double *weights, PRUint32 sizeof_weights)
{

  CmguiComputedField *source_field_cast = NULL;
  if ((NULL == source_field)
    || (NS_OK != source_field->QueryInterface(NS_GET_IID(CmguiComputedField), (void **)&source_field_cast))
    || (NULL == source_field_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }

  float *float_weights = (float *)NS_Alloc(sizeof(float) * sizeof_weights);
	unsigned int i;
	for (i = 0 ; i < sizeof_weights ; i++)
	{
		float_weights[i] = weights[i];
	}

	if ( !(Cmiss_computed_field_set_type_sum_components(impl, source_field_cast->impl, float_weights)) )
	{
		NS_Free(float_weights);
		return NS_ERROR_UNEXPECTED;
	}

	NS_Free(float_weights);
  return NS_OK;
	
}


/* void getTypeThresholdImageFilter (out CmguiIComputedField source_field, out unsigned long threshold_mode, out double outside_value, out double below_value, out double above_value); */
NS_IMETHODIMP CmguiComputedField::GetTypeThresholdImageFilter(CmguiIComputedField **source_field, PRUint32 *threshold_mode, double *outside_value, double *below_value, double *above_value)
{
  Cmiss_computed_field_id source_field_cmgui;
	if (!(Computed_field_get_type_threshold_image_filter(impl, &source_field_cmgui, (General_threshold_filter_mode*)threshold_mode, outside_value, below_value, above_value)))
  {
    return NS_ERROR_UNEXPECTED;
  }
  *source_field = new CmguiComputedField(source_field_cmgui);
  NS_ADDREF(*source_field);
  return NS_OK;
}


/* void setTypeThresholdImageFilter (in CmguiIComputedField source_field, in unsigned long threshold_mode, in double outside_value, in double below_value, in double above_value); */
NS_IMETHODIMP CmguiComputedField::SetTypeThresholdImageFilter(CmguiIComputedField *source_field, PRUint32 threshold_mode, double outside_value, double below_value, double above_value)
{
  CmguiComputedField *source_field_cast = NULL;
  if ((NULL == source_field)
    || (NS_OK != source_field->QueryInterface(NS_GET_IID(CmguiComputedField), (void **)&source_field_cast))
    || (NULL == source_field_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }
	if (!(Cmiss_computed_field_set_type_threshold_image_filter(impl, source_field_cast->impl, (General_threshold_filter_mode)threshold_mode, outside_value, below_value, above_value)))
  {
    return NS_ERROR_UNEXPECTED;
  }
  return NS_OK;
}

/* void setTypeWindowProjectionField (in CmguiIComputedField sourceField, in CmguiISceneViewer sViewer, in unsigned long projType, in CmguiICommandData cd); */
NS_IMETHODIMP CmguiComputedField::SetTypeWindowProjectionField(CmguiIComputedField *sourceField, CmguiISceneViewer *sViewer, PRUint32 projType, CmguiICommandData *cd)
{
  CmguiComputedField *sourceField_cast = NULL;
  if ((NULL == sourceField)
    || (NS_OK != sourceField->QueryInterface(NS_GET_IID(CmguiComputedField), (void **)&sourceField_cast))
    || (NULL == sourceField_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }
  CmguiSceneViewer *sViewer_cast = NULL;
  if ((NULL == sViewer)
    || (NS_OK != sViewer->QueryInterface(NS_GET_IID(CmguiSceneViewer), (void **)&sViewer_cast))
    || (NULL == sViewer_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }
  CmguiCommandData *cd_cast = NULL;
  if ((NULL == cd)
    || (NS_OK != cd->QueryInterface(NS_GET_IID(CmguiCommandData), (void **)&cd_cast))
    || (NULL == cd_cast))
  {
     return NS_ERROR_UNEXPECTED;
  }

	// get computed field manager
  struct MANAGER(Computed_field)* cfm = Cmiss_command_data_get_computed_field_manager(cd_cast->impl);
  
  // Set it to be a window projection field...
  if (!(Computed_field_set_type_window_projection(
	  impl, sourceField_cast->impl, sViewer_cast->impl, "none_ZINC_sceneviewer",
     0, (Computed_field_window_projection_type)projType, cfm)))
  {
    return NS_ERROR_UNEXPECTED;
  }

  return NS_OK;

}
