/* ***** BEGIN LICENSE BLOCK *****
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is mozCmgui.
 *
 * The Initial Developer of the Original Code is
 * Auckland Uniservices Ltd, Auckland, New Zealand.
 * Portions created by the Initial Developer are Copyright (C) 2005
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "SceneViewerSrc.h"
extern "C" {
#include "api/cmiss_scene_viewer.h"
#ifdef XP_UNIX
#include "gtk/gtk_cmiss_scene_viewer.h"
#endif
}
#ifdef XP_UNIX
#include <gtk/gtk.h>
#endif
#ifdef XP_WIN32
#include <windows.h>
#include <stdio.h>
#endif
#include "CommandDataSrc.h"
#include "zincPlugin.h"

#ifdef XP_WIN32
class CmguiSceneViewerData
{
public:
	HWND childHWND;

	CmguiSceneViewerData(HWND childHWND) : childHWND(childHWND)
	{
	}
};
#endif
#ifdef XP_UNIX
class CmguiSceneViewerData
{
public:
	CmguiCommandData *parent;

	CmguiSceneViewerData(CmguiCommandData *parent) : parent(parent)
	{
	}
};
#endif // XP_UNIX

#ifdef XP_UNIX
static void CmguiSceneViewer_gtk_unrealize_CB(GtkObject *object, gpointer scene_viewer_void)
/*******************************************************************************
LAST MODIFIED : 24 January 2007

DESCRIPTION :
Called when the gtk widget is being destroyed.  We need to clean up our reference
and if this is the last scene viewer destroy the command data while we still have
the correct OpenGL graphics context so that display lists are freed correctly.
(Alternatively we would have to not delete our display lists when destroying objects 
and rely on the graphics contexts to free them up instead).
==============================================================================*/
{
	CmguiSceneViewer *scene_viewer;

	if ((scene_viewer=(CmguiSceneViewer *)scene_viewer_void))
	{
		// Make sure this scene_viewer is current
		Cmiss_scene_viewer_redraw_now(scene_viewer->impl);

		CmguiCommandData_closing_scene_viewer_notification(scene_viewer->data->parent,
			scene_viewer);
		scene_viewer->impl = (Cmiss_scene_viewer*)NULL;
	}
	else
	{
		printf ("CmguiSceneViewer_gtk_close_CB.  Missing scene_viewer\n");
	}
} /* CmguiSceneViewer_gtk_close_CB */
#endif // XP_UNIX

CmguiSceneViewer::CmguiSceneViewer()
{
	impl = (Cmiss_scene_viewer*)NULL;
	data = (CmguiSceneViewerData*)NULL;
}

CmguiSceneViewer::~CmguiSceneViewer()
{
	if (data)
	{
		delete data;
	}
}

/* void close (); */
NS_IMETHODIMP CmguiSceneViewer::Close()
{
    return NS_OK;
}

/* void setWindowSize (in unsigned long width, in unsigned long height); */
NS_IMETHODIMP CmguiSceneViewer::SetWindowSize(PRUint32 width, PRUint32 height)
{
#ifdef WIN32
	SetWindowPos(data->childHWND, HWND_TOP, 0, 0, width, height,
	           SWP_NOZORDER | SWP_NOMOVE);
	SendMessage(data->childHWND, WM_SIZING, 0, 0);
#else
	//width = 0;
	//height = 0;
  /* XXX fixme */
#endif
  return NS_OK;
}

CmguiSceneViewer *createSceneViewerFromPlugin(
	Cmiss_scene_viewer_package *scene_viewer_package, zincIPlugin *plugin,
	CmguiCommandData *parent)
{
	Cmiss_scene_viewer_id impl;
	void *window_handle;

  zincIPluginWindow *plugin_window = NULL;
  if ((NULL == plugin)
    || (NS_OK != plugin->QueryInterface(NS_GET_IID(zincIPluginWindow),
			 (void **)&plugin_window))
    || (NULL == plugin_window))
  {
     return NULL;
  }
	
  if (NS_OK != plugin_window->GetWindow(&window_handle))
  {
	  return NULL;
  }
#ifdef XP_WIN32
	WNDCLASS srvWC;

	srvWC.style = CS_HREDRAW | CS_VREDRAW;
	srvWC.lpfnWndProc = DefWindowProc;
	srvWC.cbClsExtra = 0;
	srvWC.cbWndExtra = 0;
	srvWC.hInstance = GetModuleHandle(NULL);
	srvWC.hIcon = NULL;
	srvWC.hCursor = LoadCursor(NULL, IDC_ARROW);
	srvWC.hbrBackground = (HBRUSH)COLOR_BACKGROUND;
	srvWC.lpszMenuName = NULL;
	srvWC.lpszClassName = "cmgui_embed_wndclass";
	RegisterClass(&srvWC);

	HWND parentHWND = (HWND)handle;
	HWND childHWND = CreateWindow(
		"cmgui_embed_wndclass",
		"",
		WS_CHILD | WS_VISIBLE,
		0, 0,
		100, 100,
		parentHWND,
		NULL,
		GetModuleHandle(NULL),
		NULL
		);
	if (childHWND == NULL)
	{
      printf("createSceneViewer: CreateWindow returned a NULL pointer.\n");
      return NULL;
	}

	impl =
      CREATE(Cmiss_scene_viewer_win32)
      (
			scene_viewer_package,
			childHWND,
			CMISS_SCENE_VIEWER_BUFFERING_DOUBLE,
			CMISS_SCENE_VIEWER_STEREO_ANY_MODE,
			/*minimum_colour_buffer_depth*/5,
			/*minimum_depth_buffer_depth*/5, 
			/*minimum_accumulation_buffer_depth*/0
			);

	if (impl == NULL)
	{
		printf("CREATE(Cmiss_scene_viewer_win32) returned NULL.\n");
      return NULL;
	}

	if (!impl)
	{
      return NULL;
	}
  
	CmguiSceneViewer* cmguiSceneViewer  = new CmguiSceneViewer(impl,
		new CmguiSceneViewerData(childHWND));
	return(cmguiSceneViewer);

#elif defined XP_UNIX
	
	GdkNativeWindow xid = reinterpret_cast<GdkNativeWindow>(window_handle);
	GtkWidget *windowPlug, *sceneViewer, *packBox;

	windowPlug = gtk_plug_new(xid);
	if (windowPlug == NULL)
	{
      printf("gtk_plug_new(%u) returned NULL.\n", xid);
      return NULL;
	}

	packBox = gtk_hbox_new(FALSE, 0);
	if (packBox == NULL)
	{
      printf("gtk_hbox_new(FALSE, 0) returned NULL.\n");
      return NULL;
	}

	sceneViewer = gtk_cmiss_scene_viewer_new(scene_viewer_package);
	if (sceneViewer == NULL)
	{
      printf("gtk_cmiss_scene_viewer_new() returned NULL.\n");
      return NULL;
	}

	gtk_widget_show(sceneViewer);
	gtk_widget_show(packBox);
	gtk_container_add(GTK_CONTAINER(windowPlug), packBox);
	gtk_container_add(GTK_CONTAINER(packBox), sceneViewer);
	gtk_widget_show(windowPlug);


	impl = gtk_cmiss_scene_viewer_get_cmiss_scene_viewer
		(
			GTK_CMISS_SCENE_VIEWER(sceneViewer)
         );

	if (!impl)
	{
		return NULL;
	}

	CmguiSceneViewer* cmguiSceneViewer  = new CmguiSceneViewer(impl, 
		new CmguiSceneViewerData(parent));

	// The destroy callback happens too late, the window is already gone so
	// we cannot make it current.
	gulong gtk_close_handler_id =
		g_signal_connect (G_OBJECT(sceneViewer), "unrealize",
			G_CALLBACK(CmguiSceneViewer_gtk_unrealize_CB), (gpointer)cmguiSceneViewer);

	CmguiCommandData_opening_scene_viewer_notification(parent, cmguiSceneViewer);

	return(cmguiSceneViewer);
#endif
}

NS_IMETHODIMP CmguiSceneViewer::SetBackgroundTexturePosition(double left, double top, double width, double height)
{
  if (!Cmiss_scene_viewer_set_background_texture_info(impl,
        left, top, width, height, /*Don't undistort*/0,
		  /*Something above zero*/100))
  {
	  return NS_ERROR_UNEXPECTED;
  }
  return NS_OK;
}

NS_IMETHODIMP CmguiSceneViewer::GetFreespinTumbleAxis(float *axisX, float *axisY, float *axisZ)
{
	float axis[3];
	if (!Cmiss_scene_viewer_get_freespin_tumble_axis(impl, axis))
	{
		return NS_ERROR_UNEXPECTED;
	}
	*axisX = axis[0];
	*axisY = axis[1];
	*axisZ = axis[2];
	return NS_OK;
}

NS_IMETHODIMP CmguiSceneViewer::StartFreespin(float tumbleAxisX, float tumbleAxisY, float tumbleAxisZ, double tumbleAngle)
{
	float axis[3] = {tumbleAxisX, tumbleAxisY, tumbleAxisZ};
	if (!Cmiss_scene_viewer_start_freespin(impl, axis, tumbleAngle))
	{
		return NS_ERROR_UNEXPECTED;
	}
	return NS_OK;
}

